// polycube:  A polyomino and polycube puzzle solver.
// Copyright 2011-2018 Matthew T. Busche.
//
// This program is free software:  you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef ANGULARHEURISTIC_HPP
#define ANGULARHEURISTIC_HPP

#include <math.h>
#include "OrderingHeuristic.hpp"
#include "Constants.hpp"

class AngularHeuristic : public OrderingHeuristic
{
    private:
        double initialAngle;
        double xc;
        double yc;
        bool   reverse;

    public:
        AngularHeuristic(double initialAngle, double xc, double yc, bool reverse = false);
        double gridPointValue(const GridPoint* gp) const override;
        virtual std::ostream& print(std::ostream& os) const override;
};

inline AngularHeuristic::AngularHeuristic(double initialAngle, double xc, double yc, bool reverse)
    :
        initialAngle(initialAngle), xc(xc), yc(yc), reverse(reverse)
{
}

inline double AngularHeuristic::gridPointValue(const GridPoint* gp) const
{
    double dx = gp->getX() - xc;
    double dy = gp->getY() - yc;
    double angle = atan2(dy, dx) - initialAngle;
    if(reverse)
        angle = -angle;
    if(angle >= TAU || angle < 0)
        angle -= floor(angle / (TAU)) * TAU;

    // Adding a small linear factor of the gridpoint id as a klunky way of
    // breaking ties.  Not really important, I just like a predictable result
    // from the sort/prioritization.
    //
    return angle + gp->id / 1e7;
}

inline std::ostream& AngularHeuristic::print(std::ostream& os) const
{
    return os << (reverse ? "A" : "a") << "@" << initialAngle * 360 / TAU << "," <<
        xc << "," << yc;
}

#endif
