// polycube:  A polyomino and polycube puzzle solver.
// Copyright 2011-2018 Matthew T. Busche.
//
// This program is free software:  you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "OrderingEntity.hpp"
#include "OrderingHeuristic.hpp"
#include "makeArray.hpp"
#include "Image.hpp"
#include "PolyPerf.hpp"

OrderingEntity::OrderingEntity(
        OrderingHeuristic*        oh,
        const DlxHead*            dlxRoot,
        int                       numShapes,
        const int*                remainingShapeList,
        GridPoint const * const * filaGrid,
        bool                      nof,
        bool                      need6Bit)
    :
        oh                 (oh),
        dlxRoot            (dlxRoot),
        numShapes          (numShapes),
        remainingShapeList (remainingShapeList),
        filaGrid           (filaGrid),
        nof                (nof),
        minIlsi            (nof ? need6Bit ? 0 : 56 : 0),
        maxIlsi            (nof ? 64 : 1),
        numFilaGridPoints  (0),
        filaLoadMeter      (NULL)
{
    int numGridPoints = maxFilaGridPoints;
    this->all         = makeArray3<std::vector<Image*> >(numGridPoints, (maxIlsi-minIlsi), numShapes, 0, minIlsi, 0);
    filaLoadMeter     = PolyPerf::getInstance()->getMeter("filaLoad", "filaInit");
}


OrderingEntity::~OrderingEntity()
{
    deleteArray3(all, 0, minIlsi, 0);
}


void OrderingEntity::initPriorityOccupancy(
        gridbitfield_t* priorityOccupancy)
{
    // Preload selectionOrder with all available fila grid point bit numbers.
    //
    for(int g = 0; g < numFilaGridPoints; ++g)
        selectionOrder[g] = g;

    oh->initSelectionOrder(numFilaGridPoints, filaGrid, selectionOrder);

    gridbitfield_t po = 0;
    for(int g = 0; g < numFilaGridPoints; ++g)
    {
        int b = selectionOrder[g];
        priorityOccupancy[b] = po;
        po |= filaGrid[b]->bit;
    }
}


void OrderingEntity::loadImages(
        int nrp,
        int numFilaGridPoints,
        const gridbitfield_t neighborOccupancy[][64])
{
    PerformanceMeasurement measure(filaLoadMeter);

    // ilsiMask[g] is a length 6 bit mask (using only bits 0 through 5) with 0
    // bits corresponding to nearest neighbors that can possibly have holes,
    // and 1 bits everywhere else.  Equivalently, ilsiMask[g] has a 1 bit for
    // bits 0 through 5 that correspond to nearest neighbors that are either
    // guaranteed occupied at the time the ordering heuristic selects the
    // cell, or correspond to non-existent neighbors.  So if, for example, the
    // neighbor of g in the -x direction does not exist (because g is a
    // left-most cell in the puzzle), then bit 5 (which is always the ILSI bit
    // for the neighbor in the -x direction) will be set.
    //
    // The only possible ilsi's will have the property (ilsi & ilsiMask[g] ==
    // ilsiMask[g]) which can be used as a guide to skip the population of
    // image lists that will never be used.  Furthermore, the appropriate
    // getIlsi function for gridpoint g is given by getIlsiMap[ilsiMask[g]].
    //
    unsigned ilsiMask[maxFilaGridPoints];

    this->numFilaGridPoints = numFilaGridPoints;

    for(int i = 0; i <= nrp+1; ++i)
        lastSelectionIndex[i] = -1;

    // If this is a fixedOrder heuristic, initialize the priorityOccupancy bit
    // mask for each GridPoint.
    // 
    gridbitfield_t priorityOccupancy[maxFilaGridPoints];
    bool pof = oh->isFixedOrder();
    if(pof)
        initPriorityOccupancy(priorityOccupancy);

    // Initialize getIlsi, and ilsiMask.
    //
    for(int g = 0; g < numFilaGridPoints; ++g)
    {
        const GridPoint* gp = filaGrid[g];
        if(nof)
        {
            //std::cout << "gp->bitNum=" << gp->bitNum;
            //std::cout << "\npriorityOcc[xx]  =" << Bits<uint64_t>(priorityOccupancy[gp->bitNum]);
            //std::cout << "\ngp->neighborMask =" << Bits<uint64_t>(gp->neighborMask);
            //std::cout << "\ngp->nbr_x        =" << Bits<uint64_t>(gp->nbr_x);
            //std::cout << "\ngp->nbr_y        =" << Bits<uint64_t>(gp->nbr_y);
            gridbitfield_t possibleHoles;
            if(pof)
                possibleHoles = (gp->neighborMask & (~priorityOccupancy[gp->bitNum]));
            else
                possibleHoles = gp->neighborMask;
            //std::cout << "\ngp->nbr_y        =" << Bits<uint64_t>(gp->nbr_y);
            //std::cout << "\npossibleHoles    =" << Bits<uint64_t>(possibleHoles);
            ilsiMask[gp->bitNum] = safeIlsi(gp, ~possibleHoles);
            //std::cout << "\nilsiMask         =" << Bits<uint64_t>(ilsiMask[gp->bitNum]) << std::endl;
            getIlsi[gp->bitNum]  = getIlsiMap[ilsiMask[gp->bitNum]];
        }
        else
            getIlsi[gp->bitNum]  = NULL;  // Shouldn't be using this if NOF is disabled.
    }

    // For each remaining Shape...
    //
    for(const DlxHead* head = (const DlxHead*) dlxRoot->left;
        !head->isGridPoint(); head = (const DlxHead*) head->left)
    {
        // ...for each image of that shape...
        //
        for(const DlxNode* node = head->down; node != head; node = node->down)
        {
            Image* image = node->image;

            for(DlxNode* n = node->right; n != node; n = n->right)
            {
                const GridPoint* gp = (const GridPoint*) n->head;

                if(nof)
                {
                    if(pof)
                    {
                        gridbitfield_t po = priorityOccupancy[gp->bitNum];
                        for(int ilsi = minIlsi; ilsi < maxIlsi; ++ilsi)
                        {
                            // If the ILSI has zeros for neighbors that don't
                            // exist, or neighbors that must be filled due to
                            // priority occupancy, then this ilsi can not
                            // possibly come up during algorithm execution so
                            // it's pointless to load the list.  Just skip it.
                            //
                            if((ilsi & ilsiMask[gp->bitNum]) == ilsiMask[gp->bitNum])
                            {
                                // Otherwise, if the image doesn't conflict with
                                // the neighbor occupancy associated with this
                                // ilsi at this gridpoint, and doesn't conflict
                                // with the priority occupancy for this grid
                                // point, then load the image to thie list.
                                //
                                if((image->layoutMask & (neighborOccupancy[gp->bitNum][ilsi] | po)) == 0)
                                    all[gp->bitNum][ilsi][image->shape.id].push_back(image);
                            }
                        }
                    }
                    else
                    {
                        for(int ilsi = minIlsi; ilsi < maxIlsi; ++ilsi)
                        {
                            if((ilsi & ilsiMask[gp->bitNum]) != ilsiMask[gp->bitNum])
                                continue;

                            if((image->layoutMask & neighborOccupancy[gp->bitNum][ilsi]) == 0)
                                all[gp->bitNum][ilsi][image->shape.id].push_back(image);
                        }
                    }
                }
                else
                {
                    if(!pof || (image->layoutMask & priorityOccupancy[gp->bitNum]) == 0)
                        all[gp->bitNum][0][image->shape.id].push_back(image);
                }
            }
        }
    }
}


void OrderingEntity::clearImages()
{
    for(const DlxHead* head = (const DlxHead*) dlxRoot->left;
        !head->isGridPoint(); head = (const DlxHead*) head->left)
    {
        for(int g = 0; g < numFilaGridPoints; ++g)
            for(int ilsi = minIlsi; ilsi < maxIlsi; ++ilsi)
                all[g][ilsi][((const Shape*) head)->id].clear();
    }

    numFilaGridPoints = 0;
}
